! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful,
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************

! ***********************************************************************
!     x_ctrl's:
!     x_ctrl(1): alpha
!     x_ctrl(19): initial mass
!     x_ctrl(20): total accretion mass in sink
!     x_ctrl(21): central h1 shutdown percentage
!     x_ctrl(5): Lacc_out
!     x_ctrl(6): Lacc_in
!     s% x_logical_ctrl(5): if true do retry (dt = dt_old/2)
!
! ***********************************************************************


!! ---------------------------------------------------------------------------
!! ---------------------------------------------------------------------------

      module mod_ramses_other_adjust_mdot

      implicit none

      logical :: table_loaded = .false.
      logical :: do_retry = .false.
      logical :: first_error = .true.

      type acc_table_t
      character(len=255) :: name
      integer            :: n
      real(kind=8)       :: min_time, max_time
      real(kind=8), allocatable, dimension(:) :: time, rate
      end type

      type(acc_table_t), target :: acc_table

      real(kind=8), parameter :: secyer = 3.1558149984d7 ! seconds per year
      real(kind=8), parameter :: msun   = 1.9892e33      ! solar mass (g)
      real(kind=8) :: mtotal

      contains

            subroutine load_acc_table
                  implicit none
                  integer :: i
                  acc_table% name = 'sink.dat'
                  open(unit=57,file=acc_table% name, &
                  form='unformatted',status='old')

                  read(57) acc_table% n
                  allocate(acc_table% time(acc_table% n), &
                  acc_table% rate(acc_table% n))
                  !print *, "Read accretion rates. n=", acc_table% n
                  read(57) acc_table% time
                  read(57) acc_table% rate
                  !print*, "Time: ", acc_table% time
                  !print*, "Acc rate: ", acc_table% rate

                  close(57)

                  mtotal = acc_table% time(1) * acc_table% rate(1) * 1d6
                  do i = 2, acc_table% n
                    mtotal = mtotal + (acc_table% time(i) - acc_table% time(i-1))*acc_table% rate(i)*1d6
                  enddo

                  ! convert time to seconds
                  acc_table% time = acc_table% time * 1e6 * secyer  ! Time is in Myrs

                  acc_table% min_time = acc_table% time(1)
                  acc_table% max_time = acc_table% time(acc_table% n)


            end subroutine load_acc_table

            function acc_rate(tstart, tend)
                  use star_def
                  implicit none
                  real(kind=8) :: tstart, tend, acc_rate, lb, mp, ub
                  real(kind=8), pointer, dimension(:) :: t, acc
                  real(kind=8) :: epsilon, mdot_max, mdot_min, epsilon_limit
                  real(kind=8) :: lower_side, upper_side
                  type (star_info), pointer :: s
                  integer, save:: ilast=1, ioff=1
                  integer      :: i, idn, iup, n
                  integer :: ilast_old, ioff_old

                  if (tend   < acc_table% min_time .or. &
                  tstart > acc_table% max_time ) then
                  acc_rate = 0.0
                  return
                  endif

                  n   =  acc_table% n
                  t   => acc_table% time
                  acc => acc_table% rate

                  ilast_old = ilast
                  ioff_old = ioff

                  ! Select lower bound of accretion table
                  idn = ilast
                  do while (t(idn) > tstart)
                  idn = idn - 1
                  if (idn == 0) exit
                  enddo
                  if (idn < n) then
                  do while (t(idn + 1) < tstart)
                  idn = idn + 1
                  if (idn == n) exit
                  enddo
                  endif
                  idn = max(idn,1)
                  ! Select upper bound of accretion table
                  iup = min(ilast + ioff, n)
                  do while (t(iup-1) > tend)
                  iup = iup - 1
                  if (iup == 1) exit
                  enddo
                  do while (t(iup) < tend)
                  iup = iup + 1
                  if (iup == n+1) exit
                  enddo
                  iup = min(iup,n)

                  ilast = iup-1
                  ioff = iup - idn
                  ! For epsilon calculation
                  mdot_max = 0.0
                  mdot_min = 1e50
                  ! Estimate rate as linear sum
                  acc_rate = 0.0
                  do i=idn, iup-1
                    lb = max(t(i),tstart)
                    mp = max(min(0.5*(t(i)+t(i+1)),tend),tstart)
                    ub = min(t(i+1),tend)

                    lower_side = (mp-lb)
                    upper_side = (ub-mp)
                    acc_rate = acc_rate + lower_side*acc(i)+upper_side*acc(i+1)
                    if (lower_side > 0.0) then
                      mdot_max = max(mdot_max, acc(i))
                      mdot_min = min(mdot_min, acc(i))
                    endif
                  enddo
                  if (upper_side > 0.0) then
                    mdot_max = max(mdot_max, acc(iup))
                    mdot_min = min(mdot_min, acc(iup))
                  endif

                  ! Final interpolated acc. rate
                  acc_rate = acc_rate / (tend - tstart)

                  ! Do not interpolate more then epsilon limit:
                  epsilon_limit = 0.30
                  epsilon = max(abs(mdot_max-acc_rate), abs(mdot_min-acc_rate))/max(acc_rate,1e-10)
                  write(*,*) "epsilon: ", epsilon
                  if (ioff > 2) then
                      if (epsilon > epsilon_limit) then
                          if (first_error) then
                            first_error = .false.
                          else
                            ! Do retry - MESA will retry with dt_new = dt/2
                            do_retry = .true.
                            ilast = ilast_old
                            ioff = ioff_old
                          endif
                      else
                          do_retry = .false.
                      endif
                  else
                      do_retry = .false.
                  endif


            end function acc_rate

      ! set use_other_adjust_mdot = .true. to enable this.
      ! your routine will be called after winds and before mass
      ! adjustment


      end module mod_ramses_other_adjust_mdot

!! ---------------------------------------------------------------------------
!! ---------------------------------------------------------------------------

      module run_star_extras

      use star_lib
      use star_def
      use const_def
      use crlibm_lib

      implicit none

      ! these routines are called by the standard run_star check_model
      contains

      !include 'standard_run_star_extras.inc'

        subroutine extras_controls(id, ierr)
           integer, intent(in) :: id
           integer, intent(out) :: ierr
           type (star_info), pointer :: s
           ierr = 0
           call star_ptr(id, s, ierr)
           if (ierr /= 0) return

           s% extras_startup => extras_startup
           s% extras_check_model => extras_check_model
           !s% extras_finish_step => extras_finish_step
           !s% extras_after_evolve => extras_after_evolve
           s% how_many_extra_history_columns => how_many_extra_history_columns
           s% data_for_extra_history_columns => data_for_extra_history_columns
           !s% how_many_extra_profile_columns => how_many_extra_profile_columns
           !s% data_for_extra_profile_columns => data_for_extra_profile_columns
           !s% job% warn_run_star_extras =.false.


        	 !New routines to add
        	 s% other_energy => accretion_other_energy
           s% other_adjust_mdot => other_ramses_adjust_mdot
           s% job% warn_run_star_extras =.false.

        end subroutine extras_controls

        integer function extras_startup(id, restart, ierr)
           integer, intent(in) :: id
           logical, intent(in) :: restart
           integer, intent(out) :: ierr

           type (star_info), pointer :: s
           ierr = 0
           call star_ptr(id, s, ierr)
           if (ierr /= 0) return
           extras_startup = 0

           !!! ------------------------------
           s% x_ctrl(19) = 0.0029!s% mstar
           !!! ------------------------------

        end function extras_startup

        subroutine other_ramses_adjust_mdot(id, ierr)
            use mod_ramses_other_adjust_mdot
            use star_def
            implicit none
            integer, intent(in) :: id
            integer, intent(out) :: ierr
            type(star_info), pointer :: s
            ierr = 0
            call star_ptr(id, s, ierr) ! retrieve star id

            if (.not. table_loaded) then
              call load_acc_table
              table_loaded = .true.
            endif

            s% mstar_dot = acc_rate(s% time_old, s% time_old + s% dt)*Msun/secyer
            ! Timestep interpolation limiting
            if (do_retry .and. s% mstar_dot > 0.0) then
              s% x_logical_ctrl(5) = .true.
            else
              s% x_logical_ctrl(5) = .false.
            endif

            s% x_ctrl(20) = mtotal * msun
            write(*,*) "Adjust_mdot: ", s% mstar_dot

        end subroutine other_ramses_adjust_mdot

        subroutine accretion_other_energy(id, ierr)

            use crlibm_lib, only: safe_log10_cr
            use star_def


            implicit none

            type (star_info), pointer :: s

            integer, intent(in) :: id
            integer, intent(out) :: ierr
            real(dp) :: mass, mdot, mdot_msun, alpha, radius, mcumul
            real(dp) :: Lacc, extra_heat_var
            real(dp), parameter :: mdot_break = 6.2e-6
            real(dp), parameter :: alpha_high = 0.5
            real(dp), parameter :: alpha_low = 0.005
            real(dp), parameter :: delta_break = 5.95e-6
            real(dp) :: numerator, denominator
            real(dp), parameter :: secyer = 3.1558149984d7 ! seconds per year
            real(dp), parameter :: msun   = 1.9892e33      ! solar mass (g)
            integer :: k
            ierr  = 0

            call star_ptr(id, s, ierr)
            if (ierr /= 0) then ! OOPS
                return
            end if

            mass = s% mstar
            mdot = s% mstar_dot

            if (mdot == 0.) return

            radius = s% r(1)
            ! ----------------------------------------------------------
            ! x_ctrl(1) is the default alpha parameter from inlist.
            !alpha = s% x_ctrl(1)

            ! ----------------------------------------------------------
            ! Time varying alpha, alpha = 0.25 if mdot > 1e-5
            ! crossover for alpha in range mdot [1e-5, 1e-6]
            ! settle at alpha = 0.01 when mdot < 1e-6
      	    mdot_msun = mdot*(secyer/msun)
    	      numerator = alpha_high * exp(mdot_msun/delta_break) + alpha_low * exp(mdot_break/delta_break)
    	      denominator = exp(mdot_break/delta_break) + exp(mdot_msun/delta_break)

    	      alpha = numerator / denominator

            mcumul = 0.
            k = 0
            Lacc = 0.5 * (s% cgrav(1) * (mdot) * (mass)) / radius
            extra_heat_var = Lacc * alpha
            s% x_ctrl(6) = extra_heat_var
            s% x_ctrl(5) = Lacc * (1-alpha)
            do while ((mcumul < mdot*s% dt .or. extra_heat_var / max(mcumul,1.) > 1e4) .and. mcumul < 0.8 * mass)
              k = k + 1
              mcumul = mcumul + s% dm(k)
            enddo

            extra_heat_var = Lacc * alpha / mcumul
            s% extra_heat(1:k) = extra_heat_var

            !write(*,*) "k = ", k
            !write(*,*) "Radius(k): " , s% r(k)
            write(*,*) "Ratio, r(k) / r(1), Heat mass / mass : " , s% r(k) / radius, mcumul / mass
            write(*,*) "Extra heat: " , extra_heat_var

        end subroutine accretion_other_energy

        integer function how_many_extra_history_columns(id, id_extra)
        integer, intent(in) :: id, id_extra
        integer :: ierr
        type (star_info), pointer :: s
        ierr = 0
        call star_ptr(id, s, ierr)
        if (ierr /= 0) return
        how_many_extra_history_columns = 3
        end function how_many_extra_history_columns


        subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
        type (star_info), pointer :: s
        integer, intent(in) :: id, id_extra, n
        character (len=maxlen_history_column_name) :: names(n)
        real(dp) :: vals(n)
        integer, intent(out) :: ierr
        real(dp) :: mdot
        ierr  = 0

        call star_ptr(id, s, ierr)
        if (ierr /= 0) then ! OOPS
            return
        end if

        mdot = s% mstar_dot

        if (mdot == 0.) then
          s% x_ctrl(5) = 0.0
          s% x_ctrl(6) = 0.0
        endif

        names(1) = 'extra_lum'
        vals(1) = s% x_ctrl(5)

        names(2) = 'extra_heat'
        vals(2) = s% x_ctrl(6)

        names(3) = 'mdot'
        vals(3) = mdot

        ierr = 0

        !note: do NOT add the extras names to history_columns.list
        ! the history_columns.list is only for the built-in log column options.
        ! it must not include the new column names you are adding here.


        end subroutine data_for_extra_history_columns

        integer function extras_check_model(id, id_extra)
           integer, intent(in) :: id, id_extra
           integer :: ierr, mod_cnt = 0
           logical :: acc_done
           type (star_info), pointer :: s
           ierr = 0
           call star_ptr(id, s, ierr)
           if (ierr /= 0) return
           extras_check_model = keep_going

           if (s% x_logical_ctrl(5)) then
             extras_check_model = retry
             write(*,*) "DOING RETRY DUE TO EPSILON LIMIT"
           endif

           write(*,*) "mass percentage: ", (s% mstar/(s% x_ctrl(20) + s% x_ctrl(19)))*100.0
           write(*,*) "mass goal: ", (s% x_ctrl(20) + s%x_ctrl(19))/msun

           if (s% mstar >= (s% x_ctrl(20) + s% x_ctrl(19))*0.985) then
             acc_done = .true.
             mod_cnt = mod_cnt + 1
             write(*,*) "mod_cnt: ", mod_cnt
           endif

           if ((acc_done .eqv. .true.) .and. (mod_cnt >= 25)) then
             if (s% center_h1 < s% x_ctrl(21)) then
               extras_check_model = terminate
               s% termination_code = t_xtra1
               termination_code_str(t_xtra1) = 'central_h1 < x_ctrl(21)'
               return
             else
               continue
            endif
           else
             continue
           endif


           ! by default, indicate where (in the code) MESA terminated
           if (extras_check_model == terminate) s% termination_code = t_extras_check_model

        end function extras_check_model


      end module run_star_extras
